;; Generated package description from cl-print.el  -*- no-byte-compile: t -*-
(define-package "cl-print" "1.0.0.20210101.91356" "CL-style generic printing" '((emacs "25")) :url "https://elpa.gnu.org/packages/cl-print.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
